/*
- 1.0
:  
*/
//-----------------------------------------------------------

#include "XZField.h"

//-----------------------------------------------------------

const char rows = 3; //   
const char cols = 3; //   
const int recursionDepth = 5; //  ( ,
							  //  )
/* 2      
 .       ,
.     
,  -     ..   .*/
const int rowInc[] = {0, -1, -1, -1,  0,  1, 1, 1};
const int colInc[] = {1,  1,  0, -1, -1, -1, 0, 1};
/*     ,   
   .    
    rowInc  colInc. , 
      .*/
const int directions[4][2] = {{0, 4},
							  {1, 5},
							  {2, 6},
							  {3, 7}};
//  ( )   
//   
const int winNumber = 3;

//-----------------------------------------------------------

/*     possibleMoves ,
  .    possibleMoves
    data. ..   
row = 1, col = 2   5 (  33).
     ,      ,  
,      .
          i- , 
     "10 ^ (recursionDepth - recNumber)".
          i- , 
     "(-1) * (10 ^ (recursionDepth - recNumber))".
f -       ,
recNumber -  ,
startRow, startCol -    ( ),
possibleMoves - ,     */

void XZField::analyseMove(XZField *f, int recNumber, int startRow, int startCol,
						  int *posMoves)
{
	//------------  -------------
	//  ,       - 
	if(recNumber >= recursionDepth)
		return;

	//   
	int w;
	if(f->isGameOver(w))
	{
		//    ,    
		//  posMoves
		if(recNumber % 2 == 0)
		{
			posMoves[startRow * cols + startCol] +=
						(int)pow(10, recursionDepth - recNumber);
		}
		else
		{
			posMoves[startRow * cols + startCol] +=
						(int)((-1) * pow(10, recursionDepth - recNumber));
		}
		return;
	}

	recNumber++;
	//     ,   
	for(int i = 0; i < rows; i++)
		for(int j = 0; j < cols; j++)
		{
			if(f->getElement(i, j) == EMPTY)
			{
				XZField tempField(*f);
				//   ,      
				if(recNumber % 2 == 0)
				{
					tempField.setElement(CROSS, i, j);
					analyseMove(&tempField, recNumber, startRow, startCol, posMoves);
				}
				//  - 
				else
				{
					tempField.setElement(ZERO, i, j);
					analyseMove(&tempField, recNumber, startRow, startCol, posMoves);
				}
			}
		}
}

//-----------------------------------------------------------

/*     .
       possibleMoves
(    analyseMove).
    true,    - false.*/

bool XZField::selectMove(int &row, int &col)
{
	//    possibleMoves,
	//    "0x80000000",   .
	int max = 0x80000000;
	for(int i = 0; i < rows * cols; i++)
	{
		if(possibleMoves[i] > max)
		{
			max = possibleMoves[i];
			row = i / cols;
			col = i % cols;
		}
	}
	if(max == 0x80000000)
		return false;
	return true;
}

//-----------------------------------------------------------

XZField::XZField() : data(NULL), possibleMoves(NULL)
{
	data = new int[rows * cols];
	for(int i = 0; i < rows * cols; i++)
	{
		data[i] = EMPTY;
	}
}

//-----------------------------------------------------------

XZField::XZField(const XZField &field) : data(NULL), possibleMoves(NULL),
			winCol(-1), winRow(-1), winDirection(-1)
{
	data = new int[rows * cols];
	for(int i = 0; i < rows * cols; i++)
	{
		data[i] = field.getElement(i / rows, i % rows);
	}
}

//-----------------------------------------------------------

XZField::~XZField()
{
	if(data != NULL)
	{
		delete [] data;
	}
	if(possibleMoves != NULL)
	{
		delete [] possibleMoves;
	}
}

//-----------------------------------------------------------

//   (type)   (row, col)
//        
//     false
//    true

bool XZField::setElement(int type, int row, int col)
{
	if(row < 0 || row >= rows || col < 0 || col >= cols)
	{
		ExceptionBase e(INCORRECTINDEX, "  ");
		throw e;
	}
	if(data[row * cols + col] != EMPTY)
		return false;
	data[row * cols + col] = type;
	return true;
}
	
//-----------------------------------------------------------

// ,    (row, col)
//        

int XZField::getElement(int row, int col) const
{
	if(row < 0 || row >= rows || col < 0 || col >= cols)
	{
		ExceptionBase e(INCORRECTINDEX, "  ");
		throw e;
	}
	return data[row * cols + col];
}

//-----------------------------------------------------------

/* true   , false -   .
	  winner   :
	  EMPTY - ,
	  CROSS - ,
	  ZERO - .*/
bool XZField::isGameOver(int &winner)
{
	//    
	int count = 0;
	for(int i = 0; i < rows; i++)
		for(int j = 0; j < cols; j++)
			if(getElement(i, j) != EMPTY)
				count++;

	int crosses = 0;
	int zeros = 0;
	//   
	for(int i = 0; i < rows; i++)
		for(int j = 0; j < cols; j++)
		{
			//      
			//   
			for(int k = 0; k < 4; k++)
			{
				crosses = 0;
				zeros = 0;
				//   
				if(getElement(i, j) == CROSS)
					crosses++;
				if(getElement(i, j) == ZERO)
					zeros++;
				/*    ,   
				   .
				nextRow0  nextCol0    
				   directions, nextRow1  nextCol1 -
				 .*/
				int nextRow0 = i + rowInc[directions[k][0]];
				int	nextCol0 = j + colInc[directions[k][0]];
				int nextRow1 = i + rowInc[directions[k][1]];
				int	nextCol1 = j + colInc[directions[k][1]];
				//   (   directions)
				while(nextRow0 >= 0 && nextRow0 < rows &&
					nextCol0 >= 0 && nextCol0 < cols)
				{
					if(getElement(nextRow0, nextCol0) == CROSS)
						crosses++;
					if(getElement(nextRow0, nextCol0) == ZERO)
						zeros++;
					nextRow0 = nextRow0 + rowInc[directions[k][0]];
					nextCol0 = nextCol0 + colInc[directions[k][0]];
				}
				//   (   directions)
				while(nextRow1 >= 0 && nextRow1 < rows &&
					nextCol1 >= 0 && nextCol1 < cols)
				{
					if(getElement(nextRow1, nextCol1) == CROSS)
						crosses++;
					if(getElement(nextRow1, nextCol1) == ZERO)
						zeros++;
					nextRow1 = nextRow1 + rowInc[directions[k][1]];
					nextCol1 = nextCol1 + colInc[directions[k][1]];
				}
				if(crosses == winNumber || zeros == winNumber)
				{
					//     
					//     
					winRow = i;
					winCol = j;
					winDirection = k;
					// 
					if(crosses == winNumber)
						winner = CROSS;
					if(zeros == winNumber)
						winner = ZERO;
					return true;
				}
			}
		}
	//    ,   
	if(count == rows * cols)
	{
		winCol = winRow = winDirection = -1;
		winner = EMPTY;
		return true;
	}

	return false;
}
//-----------------------------------------------------------

/*  .      row
 col.   ,  true,    - false.*/
bool XZField::findNextMove(int &row, int &col)
{
	//  possibleMoves  ,   
	if(possibleMoves == NULL)
		possibleMoves = new int[rows * cols];
	//  possibleMoves 
	ZeroMemory(possibleMoves, rows * cols * sizeof(int));
	//   
	for(int i = 0; i < rows; i++)
		for(int j = 0; j < cols; j++)
		{
			if(getElement(i, j) == EMPTY)
			{
				//    
				XZField f(*this);
				f.setElement(CROSS, i, j);
				analyseMove(&f, 0, i, j, possibleMoves);
			}
			else
			{
				/*        
				  possibleMoves  "0x80000000"
				(    int).*/
				possibleMoves[i * cols + j] = 0x80000000;
			}
		}
	//  
	bool result = selectMove(row, col);
	return result;
}
//-----------------------------------------------------------
