/*
- 1.0
:  
*/
//-----------------------------------------------------------

#include "XZGame.h"
#include <time.h>
#include <stdlib.h>

//-----------------------------------------------------------

//     
void XZGame::ReleaseAll()
{
	//   
	if(cross != NULL)
		delete cross;
	if(zero != NULL)
		delete zero;
	if(field != NULL)
		delete field;
	if(selection != NULL)
		delete selection;
	if(gameMenu != NULL)
		delete gameMenu;
	if(player != NULL)
		delete player;
	//  Direct3D
	if(pD3DDevice != NULL)
		pD3DDevice->Release();
	if(pD3D != NULL)
		pD3D->Release();
}

//-----------------------------------------------------------

//    
D3DXMATRIX XZGame::setUpProjectionMatrix(const float near_plane, const float far_plane,
									const float fov_horiz, const float fov_vert)
{
    float    h, w, Q;

    w = (float)1/tan(fov_horiz*0.5);  // 1/tan(x) == cot(x)
    h = (float)1/tan(fov_vert*0.5);   // 1/tan(x) == cot(x)
    Q = far_plane/(far_plane - near_plane);

    D3DXMATRIX ret;
    ZeroMemory(&ret, sizeof(ret));

    ret(0, 0) = w;
    ret(1, 1) = h;
    ret(2, 2) = Q;
    ret(3, 2) = -Q*near_plane;
    ret(2, 3) = 1;
    return ret;
}

//-----------------------------------------------------------

/*        
       areaX, areaY,
        ,      "-1",
    true -     , false -   .*/
bool XZGame::checkCursorPosition(int xPos, int yPos)
{
	//  
	h = a.y - yPos;
	// h < 0 -     
	if(h < 0)
	{
		areaX = areaY = -1;
		return false;
	}
	l = L * h / H;
	innerBase = downBase - 2 * l;
	stepX = innerBase / 3;
	//       (areaX, areaY)
	// areaY
	if(yPos > b.y)
	{
		if(yPos > ab2.y)
		{
			if(yPos > ab1.y)
			{
				if(yPos > a.y)
				{
					areaY = -1;
					return false;
				}
				else
				{
					areaY = 2;
				}
			}
			else
			{
				areaY = 1;
			}
		}
		else
		{
			areaY = 0;
		}
	}
	else
	{
		areaY = -1;
		return false;
	}
	// areaX
	if((xPos > (a.x + l)) && (xPos < (d.x - l)))
	{
		//     
		areaX = floor((xPos - (a.x + l)) / stepX);
	}
	else
	{
		//    
		areaX = -1;
		return false;
	}
	return true;
}

//-----------------------------------------------------------

//  
void XZGame::startNewGame()
{
	// AI
	if(player != NULL)
		delete player;
	// AI
	player = new XZField();
	//    
	srand(time(NULL));
	if(rand() % 2 == 0)
	{
		// 
		curState = PLAYER_MOVE;
		Render();
	}
	else
	{
		// 
		curState = COMP_MOVE;
		Render();
		//  
		int r, c;
		player->findNextMove(r, c);
		//  
		player->setElement(player->CROSS, r, c);
		//  
		curState = PLAYER_MOVE;
	}
}

//-----------------------------------------------------------

//       
void XZGame::setViewPortScalingMatrix()
{
	//,      
	D3DVIEWPORT9 viewPort;
	//    
	if(FAILED(pD3DDevice->GetViewport(&viewPort)))
	{
		// 
		ExceptionBase err(GETVIEWPORTERROR, "     ");
		throw err;
	}
	//  viewPortScaling 
	D3DXMatrixIdentity(&viewPortScaling);
	//       
	viewPortScaling(0, 0) = viewPort.Width / 2;
	viewPortScaling(1, 1) = (-1) * (float)viewPort.Height / 2;
	viewPortScaling(2, 2) = viewPort.MaxZ - viewPort.MinZ;
	viewPortScaling(3, 3) = 1;
	viewPortScaling(3, 0) = viewPort.X + viewPort.Width / 2;
	viewPortScaling(3, 1) = viewPort.Height / 2 + viewPort.Y;
	viewPortScaling(3, 2) = viewPort.MinZ;
}

//-----------------------------------------------------------

//
XZGame::XZGame(void) : cross(NULL), zero(NULL), field(NULL),
	selection(NULL), gameMenu(NULL), player(NULL), pD3D(NULL),
	pD3DDevice(NULL), curState(NEW_GAME_MENU), areaX(-1), areaY(-1)
{
	D3DXMatrixIdentity(&worldMatrix);
	//    (   
	// )
	a.w = b.w = c.w = d.w = ab1.w = ab2.w = 1.0f;
	a.y = b.y = c.y = d.y = ab1.y = ab2.y = 0.0f;
	a.x = -1.5f;
	a.z = -1.5f;
	b.x = -1.5f;
	b.z = 1.5f;
	c.x = 1.5f;
	c.z = 1.5f;
	d.x = 1.5f;
	d.z = -1.5f;
	ab1.x = -1.5f;
	ab1.z = -0.5f;
	ab2.x = -1.5f;
	ab2.z = 0.5f;
}

//-----------------------------------------------------------

//
XZGame::~XZGame(void)
{
	ReleaseAll();
}

//-----------------------------------------------------------

//    Direct3D   
void XZGame::initD3DAndGameObjects()
{
	//    
	RECT clientRect;
	GetClientRect(baseHWindow, &clientRect);
	//   Direct3D   
	dx = (float)((clientRect.right - clientRect.left)/2);
	dy = (float)((clientRect.bottom - clientRect.top)/2);
	//  
	clientWidth = dx * 2;
	clientHeight = dy * 2;

	//  Direct3D
	if((pD3D = Direct3DCreate9(D3D_SDK_VERSION)) == NULL)
	{
		// 
		ExceptionBase err(D3DINITERROR, "  Direct3D");
		throw err;
	}
	//    
	D3DPRESENT_PARAMETERS d3dpp;
	//    
	D3DDISPLAYMODE d3ddm;
	//   
	if(FAILED(pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3ddm)))
	{
		// 
		ExceptionBase err(GETDISPLAYMODEERROR, "    ");
		throw err;
	}
	//    
	ZeroMemory(&d3dpp, sizeof(d3dpp));
	//  
	d3dpp.Windowed = TRUE;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = d3ddm.Format;
	d3dpp.EnableAutoDepthStencil = TRUE; // Z-
	d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

	//  Direct3D,    -  
	if(FAILED(pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,
		baseHWindow, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &pD3DDevice)))
	{
		ExceptionBase err(D3DDEVICEERROR, "    Direct3D");
		throw err;
	}

	//     
	setViewPortScalingMatrix();

	//-----------  --------------
	//
	cross = new XZMesh(pD3DDevice, "models\\cross.x");
	//
	zero = new XZMesh(pD3DDevice, "models\\zero.x");
	//
	field = new XZMesh(pD3DDevice, "models\\grid.x");
	// 
	selection = new XZMesh(pD3DDevice, "models\\selectionPlane.x");
	//   (      ,  
	//  texturesNames   NewGameMenu.h)
	char *menuFiles[] = {"textures\\NewGameMenu.png",
		"textures\\NewGameMenu_y.png", "textures\\NewGameMenu_n.png",
		"textures\\NewGameMenuWin.png", "textures\\NewGameMenuWin_y.png",
		"textures\\NewGameMenuWin_n.png", "textures\\NewGameMenuLost.png",
		"textures\\NewGameMenuLost_y.png", "textures\\NewGameMenuLost_n.png",
		"textures\\NewGameMenuDraw.png", "textures\\NewGameMenuDraw_y.png",
		"textures\\NewGameMenuDraw_n.png"};
	//
	RECT menuRect;
	menuRect.left = (int)(clientWidth / 2 - 128);
	menuRect.right = (int)(clientWidth / 2 + 128);
	menuRect.top = (int)(clientHeight / 2 - 128);
	menuRect.bottom = (int)(clientHeight / 2 + 128);
	gameMenu = new NewGameMenu(pD3DDevice, D3DXVECTOR3(0.0f, 0.0f, 0.0f), menuFiles,
		12, 256, 256, menuRect);
}

//-----------------------------------------------------------

//    Direct3D
void XZGame::setMatricesAndRenderStates()
{
	//   
	D3DXMATRIX viewMatrix, projectionMatrix;
	D3DXVECTOR3 eye(0.0f, 3.0f, -3.0f); // 
	D3DXVECTOR3 lookAt(0.0f, 0.0f, 0.0f); //    
	D3DXVECTOR3 up(0.0f, 1.0f, 0.0f); //
	//  
	D3DXMatrixLookAtLH(&viewMatrix, &eye, &lookAt, &up);
	// 
	pD3DDevice->SetTransform(D3DTS_VIEW, &viewMatrix);

	//  
	projectionMatrix = setUpProjectionMatrix(1.0f, 10.0f, 1.0f, 0.8f);
	pD3DDevice->SetTransform(D3DTS_PROJECTION, &projectionMatrix);

	//        
	D3DXMATRIX wvps; //  (    - ,
					 //,      )
	D3DXMatrixIdentity(&wvps);
	//  
	D3DXMatrixMultiply(&wvps, &worldMatrix, &viewMatrix);
	D3DXMatrixMultiply(&wvps, &wvps, &projectionMatrix);
	D3DXMatrixMultiply(&wvps, &wvps, &viewPortScaling);
	//     
	D3DXVec4Transform(&a, &a, &wvps);
	D3DXVec4Transform(&b, &b, &wvps);
	D3DXVec4Transform(&c, &c, &wvps);
	D3DXVec4Transform(&d, &d, &wvps);
	D3DXVec4Transform(&ab1, &ab1, &wvps);
	D3DXVec4Transform(&ab2, &ab2, &wvps);
	//    4-  (     
	//  )
	a = a / a.w;
	b = b / b.w;
	c = c / c.w;
	d = d / d.w;
	ab1 = ab1 / ab1.w;
	ab2 = ab2 / ab2.w;
	//   
	downBase = d.x - a.x;
	upBase = c.x - b.x;
	L = (downBase - upBase) / 2;
	H = a.y - b.y;

	//---      ---
	D3DXVECTOR4 normal = eye - lookAt;
	D3DXMATRIX menuRot;
	D3DXMatrixRotationX(&menuRot, (-1)*D3DX_PI/4);
	D3DXMATRIX menuTrans;
	D3DXMatrixTranslation(&menuTrans, normal.x / 2, normal.y / 2, normal.z / 2);
	D3DXVECTOR4 aa(0.5f, 0.0f, 0.5f, 1.0f);
	D3DXVECTOR4 bb(-0.5f, 0.0f, -0.5f, 1.0f);
	D3DXMATRIX temp;
	D3DXMatrixIdentity(&temp);
	D3DXMatrixMultiply(&temp, &menuRot, &menuTrans);
	D3DXMatrixMultiply(&temp, &temp, &viewMatrix);
	D3DXMatrixMultiply(&temp, &temp, &projectionMatrix);
	D3DXMatrixMultiply(&temp, &temp, &viewPortScaling);
	D3DXVec4Transform(&aa, &aa, &temp);
	D3DXVec4Transform(&bb, &bb, &temp);
	aa = aa / aa.w;
	bb = bb / bb.w;
	float kx = fabs(256.0f / (aa.x - bb.x));
	float kz = fabs(256.0f / (aa.y - bb.y));
	D3DXMatrixScaling(&menuMatrix, kx, 0.0f, kz);
	D3DXMatrixMultiply(&menuMatrix, &menuMatrix, &menuRot);
	D3DXMatrixMultiply(&menuMatrix, &menuMatrix, &menuTrans);

	//------   -------
	//  ,     
	pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	// Z-
	pD3DDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
	pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);

	pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    pD3DDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCCOLOR );
    pD3DDevice->SetRenderState( D3DRS_DESTBLEND,D3DBLEND_INVSRCCOLOR );
	//----  ----
	//   
	D3DXVECTOR3 direction;
	direction = lookAt - eye;
//	direction = D3DXVECTOR3(0.0f, -1.0f, 0.0f);
	//  
	D3DLIGHT9 light;
	ZeroMemory(&light, sizeof(D3DLIGHT9));
	//    ()
	light.Type = D3DLIGHT_DIRECTIONAL;
	//   ()
	light.Diffuse.r = 1.0f;
	light.Diffuse.g = 1.0f;
	light.Diffuse.b = 1.0f;
	//   
	light.Range = 100.0f;
	//  
	light.Direction = direction;
	//     
	pD3DDevice->SetLight(0, &light);
	pD3DDevice->LightEnable(0, TRUE);

	// 
	pD3DDevice->SetRenderState(D3DRS_LIGHTING, TRUE);

	//  
	pD3DDevice->SetRenderState(D3DRS_AMBIENT, D3DXCOLOR(0.3f, 0.3f, 0.3f, 1.0f));
}

//-----------------------------------------------------------

//  
bool XZGame::MessageHandler(UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	// 
	switch(iMessage)
	{
	//  
	case WM_MOUSEMOVE:
		//  
		xScreen = LOWORD(lParam);
		yScreen = HIWORD(lParam);
		
		if(curState == PLAYER_MOVE)
		{
			//      
			bool areaSelected = checkCursorPosition(xScreen, yScreen);
			//     ...
			if(areaSelected == false)
			{
				//...
				break;
			}
		}
		if(curState == NEW_GAME_MENU || curState == NEW_GAME_MENU_WIN ||
			curState == NEW_GAME_MENU_LOST || curState == NEW_GAME_MENU_DRAW)
		{
			POINT cursorPos;
			cursorPos.x = xScreen;
			cursorPos.y = yScreen;
			//   
			int menuBaseIndex;
			switch(curState)
			{
			case NEW_GAME_MENU:
				menuBaseIndex = NewGameMenu.NEW_GAME;
				break;
			case NEW_GAME_MENU_WIN:
				menuBaseIndex = NewGameMenu.NEW_GAME_WIN;
				break;
			case NEW_GAME_MENU_LOST:
				menuBaseIndex = NewGameMenu.NEW_GAME_LOST;
				break;
			case NEW_GAME_MENU_DRAW:
				menuBaseIndex = NewGameMenu.NEW_GAME_DRAW;
				break;
			}
			//      
			int selectedButton = gameMenu->checkSelectedButtons(cursorPos);
			//      -   
			//  
			if(selectedButton == NewGameMenu.NOTSELECTED)
			{
				gameMenu->setCurrentMenu(menuBaseIndex);
			}
			//      "" -   
			//   ""
			if(selectedButton == NewGameMenu.YES)
			{
				gameMenu->setCurrentMenu(menuBaseIndex + 1);
			}
			//      "" -   
			//   ""
			if(selectedButton == NewGameMenu.NO)
			{
				gameMenu->setCurrentMenu(menuBaseIndex + 2);
			}
		}
		// 
		Render();
		break;
	//   
	case WM_LBUTTONDOWN:
		//  
		xScreen = LOWORD(lParam);
		yScreen = HIWORD(lParam);
		POINT cursorPos;
		cursorPos.x = xScreen;
		cursorPos.y = yScreen;
		//     
		if(curState == NEW_GAME_MENU || curState == NEW_GAME_MENU_WIN ||
			curState == NEW_GAME_MENU_LOST || curState == NEW_GAME_MENU_DRAW)
		{
			//      
			int selectedButton = gameMenu->checkSelectedButtons(cursorPos);
			//   ""
			if(selectedButton == NewGameMenu.YES)
			{
				//  
				startNewGame();
				Render();
				break;
			}
			//   ""
			if(selectedButton == NewGameMenu.NO)
			{
				// 
				PostQuitMessage(0);
			}
			//     
			if(selectedButton == NewGameMenu.NOTSELECTED)
			{
				//
				break;
			}
		}
		//  
		if(curState == PLAYER_MOVE)
		{
			//    
			if(checkCursorPosition(xScreen, yScreen))
			{
				// ,      
				if(player->getElement(areaY, areaX) == player->EMPTY)
				{
					// 
					player->setElement(player->ZERO, areaY, areaX);
					Render();
					//,   , ,  ,
					//  
					int winner;
					if(player->isGameOver(winner))
					{
						if(winner == player->EMPTY)
						{
							curState = NEW_GAME_MENU_DRAW;
							gameMenu->setCurrentMenu(gameMenu->NEW_GAME_DRAW);
						}
						else
						{
							curState = NEW_GAME_MENU_WIN;
							gameMenu->setCurrentMenu(gameMenu->NEW_GAME_WIN);
						}
					}
					else
					{
						//  
						curState = COMP_MOVE;
						int r, c;
						player->findNextMove(r, c);
						if(!player->setElement(player->CROSS, r, c))
						{
							ExceptionBase err(FINDMOVEERROR, "    ");
							throw err;
						}
						//,   , ,  ,
						//  
						int winner;
						if(player->isGameOver(winner))
						{
							if(winner == player->EMPTY)
							{
								curState = NEW_GAME_MENU_DRAW;
								gameMenu->setCurrentMenu(gameMenu->NEW_GAME_DRAW);
							}
							else
							{
								curState = NEW_GAME_MENU_LOST;
								gameMenu->setCurrentMenu(gameMenu->NEW_GAME_LOST);
							}
						}
						else
						{
							//  
							curState = PLAYER_MOVE;
						}
					}
				}
			}
		}
		// 
		Render();
		break;
	}
	//     
	return BaseWindowClass::MessageHandler(iMessage, wParam, lParam);
}

//-----------------------------------------------------------

//    
void XZGame::Render()
{
	// 
	pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
						D3DCOLOR_XRGB(0,0,0), 1.0f, 0.0f);
	//  
	if(SUCCEEDED(pD3DDevice->BeginScene()))
	{
		//  
		pD3DDevice->SetTransform(D3DTS_WORLD, &worldMatrix);
		// 
		field->render();
		//    (,   )
		D3DXMATRIX elementTranslate;
		for(int i = 0; i < 9; i++)
		{
			// AI   -   
			if(player == NULL)
				break;
			int element = player->getElement(i % 3, i / 3);
			if(element != player->EMPTY)
			{
				D3DXMatrixIdentity(&elementTranslate);
				elementTranslate._41 = i / 3 - 1;
				elementTranslate._43 = (-1) * (i % 3 - 1);
				// 
				D3DXMatrixMultiply(&elementTranslate, &worldMatrix, &elementTranslate);
				//  
				pD3DDevice->SetTransform(D3DTS_WORLD, &elementTranslate);
				if(element == player->CROSS)
				{
					// 
					cross->render();
				}
				if(element == player->ZERO)
				{
					// 
					zero->render();
				}
			}
		}
		if(curState == PLAYER_MOVE)
		{
			//  ,     
			if(areaX != -1 && areaY != -1)
			{
				//     
				// 
				D3DXMATRIX selectionTranslate;
				D3DXMatrixIdentity(&selectionTranslate);
				selectionTranslate._41 = areaX - 1;
				selectionTranslate._43 = (-1) * (areaY - 1);
				// 
				D3DXMatrixMultiply(&selectionTranslate, &worldMatrix, &selectionTranslate);
				//  
				pD3DDevice->SetTransform(D3DTS_WORLD, &selectionTranslate);
				//  
				selection->render();
			}
		}
		if(curState == NEW_GAME_MENU || curState == NEW_GAME_MENU_WIN
			|| curState == NEW_GAME_MENU_LOST || curState == NEW_GAME_MENU_DRAW)
		{
			pD3DDevice->SetTransform(D3DTS_WORLD, &menuMatrix);
			gameMenu->render();
		}
		pD3DDevice->EndScene();
	}
	pD3DDevice->Present(NULL, NULL, NULL, NULL);
}

//-----------------------------------------------------------
